#include "MeshLoader.h"
#define PI 3.14159265

std::vector<Node>& MeshLoader::getNodes()
{
    return Nodes;
}

std::vector<FiniteElement>& MeshLoader::getFiniteElements()
{
    return FiniteElements;
}

std::vector<BoundaryFiniteElement>& MeshLoader::getBoundaryFiniteElements()
{
    return BoundaryFiniteElements;
}

std::vector<FiniteElement> MeshLoader::findElem(int id_1, int id_2)
{
    std::vector<FiniteElement> result;
    auto   last = FiniteElements.begin();
    while (last != FiniteElements.end())
    {
        last = find_if(last, FiniteElements.end(), [id_1, id_2](const FiniteElement& elem)
            {
                return (std::find(elem.nodesId.begin(), elem.nodesId.end(), id_1) != elem.nodesId.end()
                    && std::find(elem.nodesId.begin(), elem.nodesId.end(), id_2) != elem.nodesId.end());
            });

        if (last != FiniteElements.end())
            result.push_back(*last++);
    }

    return result;
}

std::vector<FiniteElement> MeshLoader::findElem(int id_1, int id_2, int id_3)
{
    std::vector<FiniteElement> result;
    auto   last = FiniteElements.begin();
    while (last != FiniteElements.end())
    {
        last = find_if(last, FiniteElements.end(), [id_1, id_2, id_3](const FiniteElement& elem)
            {
                return (std::find(elem.nodesId.begin(), elem.nodesId.end(), id_1) != elem.nodesId.end()
                    && std::find(elem.nodesId.begin(), elem.nodesId.end(), id_2) != elem.nodesId.end()
                    && std::find(elem.nodesId.begin(), elem.nodesId.end(), id_3) != elem.nodesId.end());
            });

        if (last != FiniteElements.end())
            result.push_back(*last++);
    }

    return result;
}

std::vector<Node> MeshLoader::getNodesByBoundaryFiniteElementId(int id)
{
    std::set<Node> result;
    auto   last = BoundaryFiniteElements.begin();
    while (last != BoundaryFiniteElements.end()) {
        last = std::find_if(last, BoundaryFiniteElements.end(), [id](const BoundaryFiniteElement& elem) {
            return elem.boundaryID == id; });
        if (last != BoundaryFiniteElements.end()) {
            for (const auto& it : last->nodesId) {
                auto node_by_ID = std::find_if(Nodes.begin(), Nodes.end(), [it](const Node& p_Node) { return p_Node.nodeId == it; });
                result.insert(*node_by_ID);
            }
            ++last;
        }
    }
    return std::vector<Node>(result.begin(), result.end());
}


std::vector<FiniteElement> MeshLoader::getElementsByBoundaryId(int id) 
{
    std::set<FiniteElement> result;
    auto last = BoundaryFiniteElements.begin();
    while (last != BoundaryFiniteElements.end()) {
        last = std::find_if(last, BoundaryFiniteElements.end(), [id](const BoundaryFiniteElement& elem) {
            return elem.boundaryID == id; });
        if (last != BoundaryFiniteElements.end()) {
            auto element_by_ID = std::find_if(FiniteElements.begin(), FiniteElements.end(), [last](FiniteElement& p_FiniteElement) {
                return p_FiniteElement.elementId == last->elementId; });
            result.insert(*element_by_ID);
            ++last;
        }
    }
    return std::vector<FiniteElement>(result.begin(), result.end());
}


std::vector<BoundaryFiniteElement> MeshLoader::getsBoundaryFiniteElementByBoundaryId(int id) 
{
    std::vector<BoundaryFiniteElement> result;
    auto last = BoundaryFiniteElements.begin();
    while (last != BoundaryFiniteElements.end()) {
        last = std::find_if(last, BoundaryFiniteElements.end(), [id](const BoundaryFiniteElement& elem) {
            return elem.boundaryID == id; });                                                        
        if (last != BoundaryFiniteElements.end()) {
            result.push_back(*last);
            ++last;
        }
    }
    return result;
}


std::vector<std::set<int>> MeshLoader::getNodeNeighbours() 
{

    std::vector<std::set<int>> result(getNodes().size());

    for (auto& i : FiniteElements) {
        for (int j : i.nodesId) {
            result[j - 1].insert(i.nodesId.begin(), i.nodesId.end());
            for (std::set<int>::iterator iter = result[j - 1].begin(); iter != result[j - 1].end();)
            {
                if (*iter == j)
                {
                    iter = result[j - 1].erase(iter);
                }
                else
                {
                    ++iter;
                }
            }
        }
    }
    return result;
}

void MeshLoader::insertCenterNode() {

    std::unordered_set<Edge> edges;

    for (auto& it : FiniteElements) {
        std::vector<unsigned int> current_nodes(it.nodesId);
        for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)
            for (auto j = i + 1; j < current_nodes.end(); ++j) {
                Edge current_edge(*i, *j, 0);

                auto founded_edge = edges.find(current_edge);

                if (founded_edge == edges.cend()) {
                    Node left_node = Nodes[*i - 1];
                    Node right_node = Nodes[*j - 1];
                    int center_id_ = Nodes.rbegin()->nodeId + 1;
                    Node center_node(center_id_, (left_node.cords[0] + right_node.cords[0]) / 2,
                        (left_node.cords[1] + right_node.cords[1]) / 2, (left_node.cords[2] + right_node.cords[2]) / 2, false);

                    current_edge.updateCenter(center_id_);
                    Nodes.push_back(center_node);
                    edges.insert(current_edge);
                    it.nodesId.push_back(center_id_);
                }
                else {
                    it.nodesId.push_back(founded_edge->centerNodeId);
                }
            }
        it.nodesId.shrink_to_fit();
    }

   

    for (auto& it : BoundaryFiniteElements) {
        std::vector<unsigned int> current_nodes(it.nodesId);
        for (auto i = current_nodes.begin(); i < current_nodes.end() - 1; ++i)
            for (auto j = i + 1; j < current_nodes.end(); ++j) {
                Edge current_edge(*i, *j, 0);

                auto founded_edge = edges.find(current_edge);

                if (founded_edge != edges.cend())
                    it.nodesId.push_back(founded_edge->centerNodeId);
            }
        it.nodesId.shrink_to_fit();
    }

    BoundaryFiniteElements.shrink_to_fit();
    Nodes.shrink_to_fit();
}


std::ostream& operator<<(std::ostream& stream,MeshLoader& mesh)
{
    stream << "Nodes:\n";
    for (auto& i : mesh.getNodes())
        stream << i;

    stream << "\nFiniteElements:\n";
    for (auto& i : mesh.getFiniteElements())
        stream << i;

    stream << "\nBoundaryFiniteElements:\n";
    for (auto& i : mesh.getBoundaryFiniteElements())
        stream << i;

    return stream;
}
namespace std {
    std::ostream& operator << (std::ostream& p_out, const std::vector<unsigned int>& vec) {
        for (size_t i : std::ranges::views::iota(size_t(0), vec.size()))
            p_out << vec[i] << " ";

        p_out << std::endl;
        return p_out;
    }
}


void MeshLoader::printSmallAngles(const std::string& filepath)
{
   
    std::ofstream file(filepath);
    int counter = 1;

    std::ranges::copy_if(FiniteElements |
        std::ranges::views::transform([](const FiniteElement& elem) {
            return elem.nodesId; }),
        std::ostream_iterator<std::vector<unsigned int>>(file),
        [&](const std::vector<unsigned int>& elem)
            {
                if (counter > 10)
                {
                    return false;
                }
                std::vector<std::vector<int>> planesNodes{ {0, 1, 2}, {0, 2, 3}, {0, 1, 3}, {1, 2, 3} };
                std::vector<std::vector<double>> planesCords;
                for (auto& plane : planesNodes)
                {
                    auto& node1Cords = findNodeRanges(elem[plane[0]]);
                    auto& node2Cords = findNodeRanges(elem[plane[1]]);
                    auto& node3Cords = findNodeRanges(elem[plane[2]]);

                    // Найдем уравнение плоскости по трем точкам
                    double A = node1Cords[1] * (node2Cords[2] - node3Cords[2]) +
                        node2Cords[1] * (node3Cords[2] - node1Cords[2]) +
                        node3Cords[1] * (node1Cords[2] - node2Cords[2]);

                    double B = node1Cords[2] * (node2Cords[0] - node3Cords[0]) +
                        node2Cords[2] * (node3Cords[0] - node1Cords[0]) +
                        node3Cords[2] * (node1Cords[0] - node2Cords[0]);

                    double C = node1Cords[0] * (node2Cords[1] - node3Cords[1]) +
                        node2Cords[0] * (node3Cords[1] - node1Cords[1]) +
                        node3Cords[0] * (node1Cords[1] - node2Cords[1]);

                    planesCords.push_back({ A,B,C });
                }

                // Пройдемся по всем парам плоскостей и вычислим двугранные углы
                for (int i = 0; i < 4; i++)
                    for (int j = i + 1; j < 4; j++)
                    {
                        auto& plane1 = planesCords[i];
                        auto& plane2 = planesCords[j];

                        double numerator = abs(plane1[0] * plane2[0] +
                            plane1[1] * plane2[1] +
                            plane1[2] * plane2[2]);
                        double denominator = sqrt(plane1[0] * plane1[0] +
                            plane1[1] * plane1[1] +
                            plane1[2] * plane1[2]) *
                            sqrt(plane2[0] * plane2[0] +
                                plane2[1] * plane2[1] +
                                plane2[2] * plane2[2]);


                        if (acos(numerator / denominator) * 180 / PI < 10)
                        {
                            counter++;
                            
                            return true;
                        }
                    }
                return false;
            });
}


const std::vector<double>& MeshLoader::findNodeRanges(int id)
{
    return (std::ranges::find_if(Nodes.begin(), Nodes.end(), [id](const Node& elem)
        {
            return elem.nodeId == id;
        }))->cords;
}
